"""Test helpers for Ubiquiti airOS devices."""

from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import aiohttp
import pytest

from airos.airos8 import AirOS8
from airos.exceptions import AirOSKeyDataMissingError
from airos.helpers import DetectDeviceData, async_get_firmware_data

# pylint: disable=redefined-outer-name


@pytest.fixture
def mock_session() -> MagicMock:
    """Return a mock aiohttp ClientSession."""
    return MagicMock(spec=aiohttp.ClientSession)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    (
        "mock_response",
        "expected_fw_major",
        "expected_mac",
        "expected_hostname",
        "expected_exception",
    ),
    [
        # Success case for AirOS 8
        (
            {
                "host": {"fwversion": "v8.7.4", "hostname": "test-host-8"},
                "interfaces": [
                    {"hwaddr": "AA:BB:CC:DD:EE:FF", "ifname": "br0", "enabled": True}
                ],
            },
            8,
            "AA:BB:CC:DD:EE:FF",
            "test-host-8",
            None,
        ),
        # Success case for AirOS 6
        (
            {
                "host": {"fwversion": "v6.3.16", "hostname": "test-host-6"},
                "wireless": {"mode": "sta", "apmac": "11:22:33:44:55:66"},
                "interfaces": [
                    {"hwaddr": "11:22:33:44:55:66", "ifname": "br0", "enabled": True}
                ],
            },
            6,
            "11:22:33:44:55:66",
            "test-host-6",
            None,
        ),
        # Failure case: Missing host key
        ({"wireless": {}}, 0, "", "", AirOSKeyDataMissingError),
        # Failure case: Missing fwversion key
        (
            {"host": {"hostname": "test-host"}, "interfaces": []},
            0,
            "",
            "",
            AirOSKeyDataMissingError,
        ),
        # Failure case: Invalid fwversion value
        (
            {
                "host": {"fwversion": "not-a-number", "hostname": "test-host"},
                "interfaces": [],
            },
            0,
            "",
            "",
            AirOSKeyDataMissingError,
        ),
        # Failure case: Missing hostname key
        (
            {"host": {"fwversion": "v8.7.4"}, "interfaces": []},
            0,
            "",
            "",
            AirOSKeyDataMissingError,
        ),
        # Failure case: Missing MAC address
        (
            {"host": {"fwversion": "v8.7.4", "hostname": "test-host"}},
            0,
            "",
            "",
            AirOSKeyDataMissingError,
        ),
    ],
)
async def test_firmware_detection(
    mock_session: aiohttp.ClientSession,
    mock_response: dict[str, Any],
    expected_fw_major: int,
    expected_mac: str,
    expected_hostname: str,
    expected_exception: Any,
) -> None:
    """Test helper firmware detection."""

    mock_request_json = AsyncMock(
        side_effect=[
            {},  # First call for login()
            mock_response,  # Second call for the status() endpoint
        ]
    )

    with patch.object(AirOS8, "_request_json", new=mock_request_json):
        if expected_exception:
            with pytest.raises(expected_exception):
                await async_get_firmware_data(
                    host="192.168.1.3",
                    username="testuser",
                    password="testpassword",
                    session=mock_session,
                    use_ssl=True,
                )
        else:
            # Test the success case
            device_data: DetectDeviceData = await async_get_firmware_data(
                host="192.168.1.3",
                username="testuser",
                password="testpassword",
                session=mock_session,
                use_ssl=True,
            )

            assert device_data["fw_major"] == expected_fw_major
            assert device_data["mac"] == expected_mac
            assert device_data["hostname"] == expected_hostname



"""Additional tests for airOS8 module."""

from http.cookies import SimpleCookie
import json
from unittest.mock import AsyncMock, MagicMock, patch

import aiofiles
import aiohttp
from mashumaro.exceptions import MissingField
import pytest

from airos.airos8 import AirOS8
import airos.exceptions


@pytest.mark.skip(reason="broken, needs investigation")
@pytest.mark.asyncio
async def test_login_no_csrf_token(airos8_device: AirOS8) -> None:
    """Test login response without a CSRF token header."""
    cookie = SimpleCookie()
    cookie["AIROS_TOKEN"] = "abc"

    mock_login_response = MagicMock()
    mock_login_response.__aenter__.return_value = mock_login_response
    mock_login_response.text = AsyncMock(return_value="{}")
    mock_login_response.status = 200
    mock_login_response.cookies = cookie  # Use the SimpleCookie object
    mock_login_response.headers = {}  # Simulate missing X-CSRF-ID

    with patch.object(
        airos8_device.session, "request", return_value=mock_login_response
    ):
        # We expect a return of None as the CSRF token is missing
        await airos8_device.login()


@pytest.mark.asyncio
async def test_login_connection_error(airos8_device: AirOS8) -> None:
    """Test aiohttp ClientError during login attempt."""
    with (
        patch.object(airos8_device.session, "request", side_effect=aiohttp.ClientError),
        pytest.raises(airos.exceptions.AirOSDeviceConnectionError),
    ):
        await airos8_device.login()


# --- Tests for status() and derived_data() logic ---
@pytest.mark.asyncio
async def test_status_when_not_connected(airos8_device: AirOS8) -> None:
    """Test calling status() before a successful login."""
    airos8_device.connected = False  # Ensure connected state is false
    with pytest.raises(airos.exceptions.AirOSDeviceConnectionError):
        await airos8_device.status()


# pylint: disable=pointless-string-statement
'''
@pytest.mark.asyncio
async def test_status_non_200_response(airos8_device: AirOS8) -> None:
    """Test status() with a non-successful HTTP response."""
    airos8_device.connected = True
    mock_status_response = MagicMock()
    mock_status_response.__aenter__.return_value = mock_status_response
    mock_status_response.text = AsyncMock(return_value="Error")
    mock_status_response.status = 500  # Simulate server error

    with (
        patch.object(airos8_device.session, "request", return_value=mock_status_response),
        pytest.raises(airos.exceptions.AirOSDeviceConnectionError),
    ):
        await airos8_device.status()
'''


@pytest.mark.asyncio
async def test_status_invalid_json_response(airos8_device: AirOS8) -> None:
    """Test status() with a response that is not valid JSON."""
    airos8_device.connected = True
    mock_status_response = MagicMock()
    mock_status_response.__aenter__.return_value = mock_status_response
    mock_status_response.text = AsyncMock(return_value="This is not JSON")
    mock_status_response.status = 200

    with (
        patch.object(
            airos8_device.session, "request", return_value=mock_status_response
        ),
        pytest.raises(airos.exceptions.AirOSDataMissingError),
    ):
        await airos8_device.status()


@pytest.mark.asyncio
async def test_status_missing_interface_key_data(airos8_device: AirOS8) -> None:
    """Test status() with a response missing critical data fields."""
    airos8_device.connected = True
    # The derived_data() function is called with a mocked response
    mock_status_response = MagicMock()
    mock_status_response.__aenter__.return_value = mock_status_response
    mock_status_response.text = AsyncMock(
        return_value=json.dumps({"system": {}})
    )  # Missing 'interfaces'
    mock_status_response.status = 200

    with (
        patch.object(
            airos8_device.session, "request", return_value=mock_status_response
        ),
        pytest.raises(airos.exceptions.AirOSKeyDataMissingError),
    ):
        await airos8_device.status()


@pytest.mark.asyncio
async def test_derived_data_no_interfaces_key(airos8_device: AirOS8) -> None:
    """Test derived_data() with a response that has no 'interfaces' key."""
    # This will directly test the 'if not interfaces:' branch (line 206)
    with pytest.raises(airos.exceptions.AirOSKeyDataMissingError):
        airos8_device.derived_data({})


@pytest.mark.asyncio
async def test_derived_data_no_br0_eth0_ath0(airos8_device: AirOS8) -> None:
    """Test derived_data() with an unexpected interface list, to test the fallback logic."""
    fixture_data = {
        "interfaces": [
            {"ifname": "wan0", "enabled": True, "hwaddr": "11:22:33:44:55:66"}
        ]
    }

    adjusted_data = airos8_device.derived_data(fixture_data)
    assert adjusted_data["derived"]["mac_interface"] == "wan0"
    assert adjusted_data["derived"]["mac"] == "11:22:33:44:55:66"


# --- Tests for stakick() ---
@pytest.mark.asyncio
async def test_stakick_when_not_connected(airos8_device: AirOS8) -> None:
    """Test stakick() before a successful login."""
    airos8_device.connected = False
    with pytest.raises(airos.exceptions.AirOSDeviceConnectionError):
        await airos8_device.stakick("01:23:45:67:89:aB")


@pytest.mark.asyncio
async def test_stakick_no_mac_address(airos8_device: AirOS8) -> None:
    """Test stakick() with a None mac_address."""
    airos8_device.connected = True
    with pytest.raises(airos.exceptions.AirOSDataMissingError):
        await airos8_device.stakick(None)


@pytest.mark.skip(reason="broken, needs investigation")
@pytest.mark.asyncio
async def test_stakick_non_200_response(airos8_device: AirOS8) -> None:
    """Test stakick() with a non-successful HTTP response."""
    airos8_device.connected = True
    mock_stakick_response = MagicMock()
    mock_stakick_response.__aenter__.return_value = mock_stakick_response
    mock_stakick_response.text = AsyncMock(return_value="Error")
    mock_stakick_response.status = 500

    with patch.object(
        airos8_device.session, "request", return_value=mock_stakick_response
    ):
        assert not await airos8_device.stakick("01:23:45:67:89:aB")


@pytest.mark.asyncio
async def test_stakick_connection_error(airos8_device: AirOS8) -> None:
    """Test aiohttp ClientError during stakick."""
    airos8_device.connected = True
    with (
        patch.object(airos8_device.session, "request", side_effect=aiohttp.ClientError),
        pytest.raises(airos.exceptions.AirOSDeviceConnectionError),
    ):
        await airos8_device.stakick("01:23:45:67:89:aB")


# --- Tests for provmode() (Complete Coverage) ---
@pytest.mark.asyncio
async def test_provmode_when_not_connected(airos8_device: AirOS8) -> None:
    """Test provmode() before a successful login."""
    airos8_device.connected = False
    with pytest.raises(airos.exceptions.AirOSDeviceConnectionError):
        await airos8_device.provmode(active=True)


@pytest.mark.skip(reason="broken, needs investigation")
@pytest.mark.asyncio
async def test_provmode_activate_success(airos8_device: AirOS8) -> None:
    """Test successful activation of provisioning mode."""
    airos8_device.connected = True
    mock_provmode_response = MagicMock()
    mock_provmode_response.__aenter__.return_value = mock_provmode_response
    mock_provmode_response.status = 200
    mock_provmode_response.text = AsyncMock()
    mock_provmode_response.text.return_value = ""

    with patch.object(
        airos8_device.session, "request", return_value=mock_provmode_response
    ):
        assert await airos8_device.provmode(active=True)


@pytest.mark.skip(reason="broken, needs investigation")
@pytest.mark.asyncio
async def test_provmode_deactivate_success(airos8_device: AirOS8) -> None:
    """Test successful deactivation of provisioning mode."""
    airos8_device.connected = True
    mock_provmode_response = MagicMock()
    mock_provmode_response.__aenter__.return_value = mock_provmode_response
    mock_provmode_response.status = 200
    mock_provmode_response.text = AsyncMock()
    mock_provmode_response.text.return_value = ""

    with patch.object(
        airos8_device.session, "request", return_value=mock_provmode_response
    ):
        assert await airos8_device.provmode(active=False)


@pytest.mark.skip(reason="broken, needs investigation")
@pytest.mark.asyncio
async def test_provmode_non_200_response(airos8_device: AirOS8) -> None:
    """Test provmode() with a non-successful HTTP response."""
    airos8_device.connected = True
    mock_provmode_response = MagicMock()
    mock_provmode_response.__aenter__.return_value = mock_provmode_response
    mock_provmode_response.text = AsyncMock(return_value="Error")
    mock_provmode_response.status = 500

    with patch.object(
        airos8_device.session, "request", return_value=mock_provmode_response
    ):
        assert not await airos8_device.provmode(active=True)


@pytest.mark.asyncio
async def test_provmode_connection_error(airos8_device: AirOS8) -> None:
    """Test aiohttp ClientError during provmode."""
    airos8_device.connected = True
    with (
        patch.object(airos8_device.session, "request", side_effect=aiohttp.ClientError),
        pytest.raises(airos.exceptions.AirOSDeviceConnectionError),
    ):
        await airos8_device.provmode(active=True)


@pytest.mark.asyncio
async def test_status_missing_required_key_in_json(airos8_device: AirOS8) -> None:
    """Test status() with a response missing a key required by the dataclass."""
    airos8_device.connected = True
    # Fixture is valid JSON, but is missing the entire 'wireless' block,
    # which is a required field for the AirOS8Data dataclass.
    invalid_data = {
        "host": {"hostname": "test"},
        "interfaces": [
            {"ifname": "br0", "hwaddr": "11:22:33:44:55:66", "enabled": True}
        ],
    }

    mock_status_response = MagicMock()
    mock_status_response.__aenter__.return_value = mock_status_response
    mock_status_response.text = AsyncMock(return_value=json.dumps(invalid_data))
    mock_status_response.status = 200

    with (
        patch.object(
            airos8_device.session, "request", return_value=mock_status_response
        ),
        patch("airos.base._LOGGER.exception") as mock_log_exception,
        pytest.raises(airos.exceptions.AirOSKeyDataMissingError) as excinfo,
    ):
        await airos8_device.status()

    # Check that the specific mashumaro error is logged and caught
    mock_log_exception.assert_called_once()
    assert "Failed to deserialize AirOS data" in mock_log_exception.call_args[0][0]
    # --- MODIFICATION START ---
    # Assert that the cause of our exception is the correct type from mashumaro
    assert isinstance(excinfo.value.__cause__, MissingField)


# --- Tests for warnings() and update_check() ---
@pytest.mark.asyncio
async def test_warnings_correctly_parses_json() -> None:
    """Test that the warnings() method correctly parses a valid JSON response."""
    mock_session = MagicMock()
    airos8_device = AirOS8(
        host="http://192.168.1.3",
        username="test",
        password="test",
        session=mock_session,
    )
    airos8_device.connected = True

    mock_response = MagicMock()
    mock_response.__aenter__.return_value = mock_response
    mock_response.status = 200
    async with aiofiles.open("fixtures/warnings.json") as f:
        content = await f.read()
        mock_response_data = json.loads(content)
    mock_response.text = AsyncMock(return_value=json.dumps(mock_response_data))

    with patch.object(airos8_device.session, "request", return_value=mock_response):
        result = await airos8_device.warnings()
        assert result["isDefaultPasswd"] is False
        assert result["chAvailable"] is False


@pytest.mark.asyncio
async def test_warnings_raises_exception_on_invalid_json() -> None:
    """Test that warnings() raises an exception on invalid JSON response."""
    mock_session = MagicMock()
    airos8_device = AirOS8(
        host="http://192.168.1.3",
        username="test",
        password="test",
        session=mock_session,
    )
    airos8_device.connected = True

    mock_response = MagicMock()
    mock_response.__aenter__.return_value = mock_response
    mock_response.status = 200
    mock_response.text = AsyncMock(return_value="This is not JSON")

    with (
        patch.object(airos8_device.session, "request", return_value=mock_response),
        pytest.raises(airos.exceptions.AirOSDataMissingError),
    ):
        await airos8_device.warnings()


@pytest.mark.asyncio
async def test_update_check_correctly_parses_json() -> None:
    """Test that update_check() method correctly parses a valid JSON response."""
    mock_session = MagicMock()
    airos8_device = AirOS8(
        host="http://192.168.1.3",
        username="test",
        password="test",
        session=mock_session,
    )
    airos8_device.connected = True
    airos8_device.current_csrf_token = "mock-csrf-token"

    mock_response = MagicMock()
    mock_response.__aenter__.return_value = mock_response
    mock_response.status = 200
    async with aiofiles.open("fixtures/update_check_available.json") as f:
        content = await f.read()
        mock_response_data = json.loads(content)
    mock_response.text = AsyncMock(return_value=json.dumps(mock_response_data))

    with patch.object(airos8_device.session, "request", return_value=mock_response):
        result = await airos8_device.update_check()
        assert result["version"] == "v8.7.19"
        assert result["update"] is True


@pytest.mark.asyncio
async def test_update_check_raises_exception_on_invalid_json() -> None:
    """Test that update_check() raises an exception on invalid JSON response."""
    mock_session = MagicMock()
    airos8_device = AirOS8(
        host="http://192.168.1.3",
        username="test",
        password="test",
        session=mock_session,
    )
    airos8_device.connected = True
    airos8_device.current_csrf_token = "mock-csrf-token"

    mock_response = MagicMock()
    mock_response.__aenter__.return_value = mock_response
    mock_response.status = 200
    mock_response.text = AsyncMock(return_value="This is not JSON")

    with (
        patch.object(airos8_device.session, "request", return_value=mock_response),
        pytest.raises(airos.exceptions.AirOSDataMissingError),
    ):
        await airos8_device.update_check()


@pytest.mark.asyncio
async def test_warnings_when_not_connected() -> None:
    """Test calling warnings() before a successful login."""
    mock_session = MagicMock()
    airos8_device = AirOS8(
        host="http://192.168.1.3",
        username="test",
        password="test",
        session=mock_session,
    )
    airos8_device.connected = False  # Explicitly set connected state to false

    with pytest.raises(airos.exceptions.AirOSDeviceConnectionError):
        await airos8_device.warnings()


@pytest.mark.asyncio
async def test_update_check_when_not_connected() -> None:
    """Test calling update_check() before a successful login."""
    mock_session = MagicMock()
    airos8_device = AirOS8(
        host="http://192.168.1.3",
        username="test",
        password="test",
        session=mock_session,
    )
    airos8_device.connected = False  # Explicitly set connected state to false

    with pytest.raises(airos.exceptions.AirOSDeviceConnectionError):
        await airos8_device.update_check()



"""Tests for airos data module."""

from unittest.mock import patch

import pytest

from airos.data import Host, Wireless


@pytest.mark.asyncio
async def test_unknown_enum_values() -> None:
    """Test that unknown enum values are handled gracefully."""
    # 1. Test for Host.netrole
    host_data = {"netrole": "unsupported_role", "other_field": "value"}
    format_string = (
        "Unknown value '%s' for %s.%s. Please report at "
        "https://github.com/CoMPaTech/python-airos/issues so we can add support."
    )
    with patch("airos.data.logger.warning") as mock_warning:
        processed_host = Host.__pre_deserialize__(host_data.copy())
        # Verify the unknown value was removed
        assert "netrole" not in processed_host
        # Verify the other fields remain
        assert "other_field" in processed_host
        # Verify a warning was logged
        mock_warning.assert_called_once_with(
            format_string, "unsupported_role", "Host", "netrole"
        )

    # 2. Test for Wireless (all enums)
    wireless_data = {
        "mode": "unsupported_mode",
        "ieeemode": "unsupported_ieee",
        "security": "unsupported_security",
        "other_field": "value",
    }
    with patch("airos.data.logger.warning") as mock_warning:
        processed_wireless = Wireless.__pre_deserialize__(wireless_data.copy())
        # Verify the unknown values were removed
        assert "mode" not in processed_wireless
        assert "ieeemode" not in processed_wireless
        assert "security" not in processed_wireless
        # Verify the other field remains
        assert "other_field" in processed_wireless
        # Verify warnings were logged for each unknown enum
        assert mock_warning.call_count == 3
        mock_warning.assert_any_call(
            format_string, "unsupported_mode", "Wireless", "mode"
        )
        mock_warning.assert_any_call(
            format_string, "unsupported_ieee", "Wireless", "ieeemode"
        )
        mock_warning.assert_any_call(
            format_string, "unsupported_security", "Wireless", "security"
        )



"""Additional tests for airOS6 module."""

from http.cookies import SimpleCookie
import json
from unittest.mock import AsyncMock, MagicMock, patch

import aiohttp
from mashumaro.exceptions import MissingField
import pytest

from airos.airos6 import AirOS6
import airos.exceptions


@pytest.mark.skip(reason="broken, needs investigation")
@pytest.mark.asyncio
async def test_login_no_csrf_token(airos6_device: AirOS6) -> None:
    """Test login response without a CSRF token header."""
    cookie = SimpleCookie()
    cookie["AIROS_TOKEN"] = "abc"

    mock_login_response = MagicMock()
    mock_login_response.__aenter__.return_value = mock_login_response
    mock_login_response.text = AsyncMock(return_value="{}")
    mock_login_response.status = 200
    mock_login_response.cookies = cookie  # Use the SimpleCookie object
    mock_login_response.headers = {}  # Simulate missing X-CSRF-ID

    with patch.object(
        airos6_device.session, "request", return_value=mock_login_response
    ):
        # We expect a return of None as the CSRF token is missing
        await airos6_device.login()


@pytest.mark.asyncio
async def test_login_connection_error(airos6_device: AirOS6) -> None:
    """Test aiohttp ClientError during login attempt."""
    with (
        patch.object(airos6_device.session, "request", side_effect=aiohttp.ClientError),
        pytest.raises(airos.exceptions.AirOSDeviceConnectionError),
    ):
        await airos6_device.login()


# --- Tests for status() and derived_data() logic ---
@pytest.mark.asyncio
async def test_status_when_not_connected(airos6_device: AirOS6) -> None:
    """Test calling status() before a successful login."""
    airos6_device.connected = False  # Ensure connected state is false
    with pytest.raises(airos.exceptions.AirOSDeviceConnectionError):
        await airos6_device.status()


# pylint: disable=pointless-string-statement
'''
@pytest.mark.asyncio
async def test_status_non_200_response(airos6_device: AirOS6) -> None:
    """Test status() with a non-successful HTTP response."""
    airos6_device.connected = True
    mock_status_response = MagicMock()
    mock_status_response.__aenter__.return_value = mock_status_response
    mock_status_response.text = AsyncMock(return_value="Error")
    mock_status_response.status = 500  # Simulate server error

    with (
        patch.object(airos6_device.session, "request", return_value=mock_status_response),
        pytest.raises(airos.exceptions.AirOSDeviceConnectionError),
    ):
        await airos6_device.status()
'''


@pytest.mark.asyncio
async def test_status_invalid_json_response(airos6_device: AirOS6) -> None:
    """Test status() with a response that is not valid JSON."""
    airos6_device.connected = True
    mock_status_response = MagicMock()
    mock_status_response.__aenter__.return_value = mock_status_response
    mock_status_response.text = AsyncMock(return_value="This is not JSON")
    mock_status_response.status = 200

    with (
        patch.object(
            airos6_device.session, "request", return_value=mock_status_response
        ),
        pytest.raises(airos.exceptions.AirOSDataMissingError),
    ):
        await airos6_device.status()


@pytest.mark.asyncio
async def test_status_missing_interface_key_data(airos6_device: AirOS6) -> None:
    """Test status() with a response missing critical data fields."""
    airos6_device.connected = True
    # The derived_data() function is called with a mocked response
    mock_status_response = MagicMock()
    mock_status_response.__aenter__.return_value = mock_status_response
    mock_status_response.text = AsyncMock(
        return_value=json.dumps({"system": {}})
    )  # Missing 'interfaces'
    mock_status_response.status = 200

    with (
        patch.object(
            airos6_device.session, "request", return_value=mock_status_response
        ),
        pytest.raises(airos.exceptions.AirOSKeyDataMissingError),
    ):
        await airos6_device.status()


@pytest.mark.asyncio
async def test_derived_data_no_interfaces_key(airos6_device: AirOS6) -> None:
    """Test derived_data() with a response that has no 'interfaces' key."""
    # This will directly test the 'if not interfaces:' branch (line 206)
    with pytest.raises(airos.exceptions.AirOSKeyDataMissingError):
        airos6_device.derived_data({})


@pytest.mark.asyncio
async def test_derived_data_no_br0_eth0_ath0(airos6_device: AirOS6) -> None:
    """Test derived_data() with an unexpected interface list, to test the fallback logic."""
    fixture_data = {
        "interfaces": [
            {"ifname": "wan0", "enabled": True, "hwaddr": "11:22:33:44:55:66"}
        ]
    }

    adjusted_data = airos6_device.derived_data(fixture_data)
    assert adjusted_data["derived"]["mac_interface"] == "wan0"
    assert adjusted_data["derived"]["mac"] == "11:22:33:44:55:66"


@pytest.mark.skip(reason="broken, needs investigation")
@pytest.mark.asyncio
async def test_status_missing_required_key_in_json(airos6_device: AirOS6) -> None:
    """Test status() with a response missing a key required by the dataclass."""
    airos6_device.connected = True
    # Fixture is valid JSON, but is missing the entire 'wireless' block,
    # which is a required field for the AirOS6Data dataclass.
    invalid_data = {
        "host": {"hostname": "test"},
        "interfaces": [
            {"ifname": "br0", "hwaddr": "11:22:33:44:55:66", "enabled": True}
        ],
    }

    mock_status_response = MagicMock()
    mock_status_response.__aenter__.return_value = mock_status_response
    mock_status_response.text = AsyncMock(return_value=json.dumps(invalid_data))
    mock_status_response.status = 200

    with (
        patch.object(
            airos6_device.session, "request", return_value=mock_status_response
        ),
        patch("airos.airos6._LOGGER.exception") as mock_log_exception,
        pytest.raises(airos.exceptions.AirOSKeyDataMissingError) as excinfo,
    ):
        await airos6_device.status()

    # Check that the specific mashumaro error is logged and caught
    mock_log_exception.assert_called_once()
    assert "Failed to deserialize AirOS data" in mock_log_exception.call_args[0][0]
    # --- MODIFICATION START ---
    # Assert that the cause of our exception is the correct type from mashumaro
    assert isinstance(excinfo.value.__cause__, MissingField)



"""Unit tests for the UispAirOSProductMapper class in model_map.py."""

import re

import pytest

from airos.exceptions import AirOSMultipleMatchesFoundException
from airos.model_map import UispAirOSProductMapper


class TestUispAirOSProductMapper:
    """Unit tests for the UispAirOSProductMapper class."""

    mapper = UispAirOSProductMapper()

    def test_get_sku_by_devmodel_exact_match(self):
        """Test to return the correct SKU for a full, exact model name."""
        sku = self.mapper.get_sku_by_devmodel("Wave MLO5")
        assert sku == "Wave-MLO5"

    def test_get_sku_by_devmodel_partial_single_match(self):
        """Test to return the correct SKU for a partial model name that matches only one product."""
        sku = self.mapper.get_sku_by_devmodel("NanoBeam 5AC")
        assert sku == "NBE-5AC-GEN2"

    def test_get_sku_by_devmodel_case_insensitivity(self):
        """Test to work regardless of the case of the input model name."""
        sku = self.mapper.get_sku_by_devmodel("nanostation 5ac loco")
        assert sku == "Loco5AC"

    def test_get_sku_by_devmodel_not_found_raises_keyerror(self):
        """Test to raise KeyError when no match (exact or partial) is found."""
        with pytest.raises(
            KeyError, match="No product found for devmodel: NonExistent"
        ):
            self.mapper.get_sku_by_devmodel("NonExistent Model 123")

    def test_get_sku_by_devmodel_multiple_matches_raises_exception_dynamic(self):
        """Test to raise AirOSMultipleMatchesFoundException when partial match is ambiguous."""
        with pytest.raises(AirOSMultipleMatchesFoundException) as excinfo:
            self.mapper.get_sku_by_devmodel("LiteBeam")

        exception_message = str(excinfo.value)
        expected_matches = 4

        match = re.search(r"matched multiple \((\d+)\) products", exception_message)
        assert match is not None
        actual_matches_int = int(match.group(1))
        assert actual_matches_int == expected_matches

    def test_get_devmodel_by_sku_exact_match(self):
        """Test to return the full model name for an exact SKU."""
        model = self.mapper.get_devmodel_by_sku("Loco5AC")
        assert model == "airMAX NanoStation 5AC Loco"

    def test_get_devmodel_by_sku_not_found_raises_keyerror(self):
        """Test to raise KeyError when the exact SKU is not found."""
        with pytest.raises(KeyError, match="No product found for SKU: FAKE-SKU"):
            self.mapper.get_devmodel_by_sku("FAKE-SKU")



"""Provide mashumaro data object for AirOSData."""

from dataclasses import dataclass
from enum import Enum
import ipaddress
import logging
import re
from typing import Any

from mashumaro import DataClassDictMixin
from mashumaro.config import BaseConfig

logger = logging.getLogger(__name__)

# Regex for a standard MAC address format (e.g., 01:23:45:67:89:AB)
# This handles both colon and hyphen separators.
MAC_ADDRESS_REGEX = re.compile(r"^([0-9a-fA-F]{2}[:-]){5}([0-9a-fA-F]{2})$")

# Regex for a MAC address mask (e.g., the redacted format 00:00:00:00:89:AB)
MAC_ADDRESS_MASK_REGEX = re.compile(r"^(00:){4}[0-9a-fA-F]{2}[:-][0-9a-fA-F]{2}$")


# Helper functions
def is_mac_address(value: str) -> bool:
    """Check if a string is a valid MAC address."""
    return bool(MAC_ADDRESS_REGEX.match(value))


def is_mac_address_mask(value: str) -> bool:
    """Check if a string is a valid MAC address mask (e.g., the redacted format)."""
    return bool(MAC_ADDRESS_MASK_REGEX.match(value))


def is_ip_address(value: str) -> bool:
    """Check if a string is a valid IPv4 or IPv6 address."""
    try:
        ipaddress.ip_address(value)
    except ValueError:
        return False
    return True  # pragma: no cover


def redact_data_smart(data: dict[str, Any]) -> dict[str, Any]:
    """Recursively redacts sensitive keys in a dictionary."""
    sensitive_keys = {
        "hostname",
        "essid",
        "mac",
        "apmac",
        "hwaddr",
        "lastip",
        "ipaddr",
        "ip6addr",
        "device_id",
        "sys_id",
        "station_id",
        "platform",
    }

    def _redact(d: dict[str, Any]) -> dict[str, Any]:
        redacted_d = {}
        for k, v in d.items():
            if k in sensitive_keys:
                if isinstance(v, str) and (is_mac_address(v) or is_mac_address_mask(v)):
                    # Redact only the last part of a MAC address to a dummy value
                    redacted_d[k] = "00:11:22:33:" + v.replace("-", ":").upper()[-5:]
                elif isinstance(v, str) and is_ip_address(v):  # pragma: no cover
                    # Redact to a dummy local IP address
                    redacted_d[k] = "127.0.0.3"
                elif isinstance(v, list) and all(
                    isinstance(i, str) and is_ip_address(i) for i in v
                ):  # pragma: no cover
                    # Redact list of IPs to a dummy list
                    redacted_d[k] = ["127.0.0.3"]  # type: ignore[assignment]
                elif isinstance(v, list) and all(
                    isinstance(i, dict) and "addr" in i and is_ip_address(i["addr"])
                    for i in v
                ):  # pragma: no cover
                    # Redact list of dictionaries with IP addresses to a dummy list
                    redacted_list = []
                    for item in v:
                        redacted_item = item.copy()
                        redacted_item["addr"] = (
                            "127.0.0.3"
                            if ipaddress.ip_address(redacted_item["addr"]).version == 4
                            else "::1"
                        )
                        redacted_list.append(redacted_item)
                    redacted_d[k] = redacted_list  # type: ignore[assignment]
                else:
                    redacted_d[k] = "REDACTED"
            elif isinstance(v, dict):
                redacted_d[k] = _redact(v)  # type: ignore[assignment]
            elif isinstance(v, list):
                redacted_d[k] = [
                    _redact(item) if isinstance(item, dict) else item for item in v
                ]  # type: ignore[assignment]
            else:
                redacted_d[k] = v
        return redacted_d

    return _redact(data)


# Data class start


class AirOSDataClass(DataClassDictMixin):
    """A base class for all mashumaro dataclasses."""


@dataclass
class AirOSDataBaseClass(AirOSDataClass):
    """Base class for all AirOS data models."""

    class Config(BaseConfig):
        """Create base class for multiple version support."""

        alias_generator = str.upper


def _check_and_log_unknown_enum_value(
    data_dict: dict[str, Any],
    key: str,
    enum_class: type[Enum],
    dataclass_name: str,
    field_name: str,
) -> None:
    """Clean unsupported parameters with logging."""
    value = data_dict.get(key)
    if value is not None and isinstance(value, str):
        if value not in [e.value for e in enum_class]:
            logger.warning(
                "Unknown value '%s' for %s.%s. Please report at "
                "https://github.com/CoMPaTech/python-airos/issues so we can add support.",
                value,
                dataclass_name,
                field_name,
            )
            del data_dict[key]


class IeeeMode(Enum):
    """Enum definition."""

    AUTO = "AUTO"
    _11ACVHT80 = "11ACVHT80"  # On a NanoStation
    _11ACVHT40 = "11ACVHT40"
    _11ACVHT20 = "11ACVHT20"  # On a LiteBeam
    # More to be added when known


class DerivedWirelessRole(Enum):
    """Enum definition."""

    STATION = "station"
    ACCESS_POINT = "access_point"


class DerivedWirelessMode(Enum):
    """Enum definition."""

    PTP = "point_to_point"
    PTMP = "point_to_multipoint"


class WirelessMode(Enum):
    """Enum definition."""

    PTMP_ACCESSPOINT = "ap-ptmp"
    PTMP_STATION = "sta-ptmp"
    PTP_ACCESSPOINT = "ap-ptp"
    PTP_STATION = "sta-ptp"
    UNKNOWN = "unknown"  # Reported on v8.7.18 NanoBeam 5AC for remote.mode
    # More to be added when known


class Wireless6Mode(Enum):
    """Enum definition."""

    STATION = "sta"
    ACCESSPOINT = "ap"
    # More to be added when known


class Security(Enum):
    """Enum definition."""

    WPA2 = "WPA2"
    # More to be added when known


class NetRole(Enum):
    """Enum definition."""

    BRIDGE = "bridge"
    ROUTER = "router"
    # More to be added when known


@dataclass
class ChainName(AirOSDataClass):
    """Leaf definition."""

    number: int
    name: str


@dataclass
class Host(AirOSDataClass):
    """Leaf definition."""

    hostname: str
    uptime: int
    power_time: int
    time: str
    timestamp: int
    fwversion: str
    devmodel: str
    netrole: NetRole
    loadavg: float | int | None
    totalram: int
    freeram: int
    temperature: int
    cpuload: float | int | None
    device_id: str
    height: int | None  # Reported none on LiteBeam 5AC

    @classmethod
    def __pre_deserialize__(cls, d: dict[str, Any]) -> dict[str, Any]:
        """Pre-deserialize hook for Host."""
        _check_and_log_unknown_enum_value(d, "netrole", NetRole, "Host", "netrole")
        return d


@dataclass
class Host6(AirOSDataClass):
    """Leaf definition."""

    hostname: str
    uptime: int
    fwversion: str
    fwprefix: str
    devmodel: str
    netrole: NetRole
    totalram: int
    freeram: int
    cpuload: float | int | None

    @classmethod
    def __pre_deserialize__(cls, d: dict[str, Any]) -> dict[str, Any]:
        """Pre-deserialize hook for Host."""
        _check_and_log_unknown_enum_value(d, "netrole", NetRole, "Host", "netrole")
        return d


@dataclass
class Services(AirOSDataClass):
    """Leaf definition."""

    dhcpc: bool
    dhcpd: bool
    dhcp6d_stateful: bool
    pppoe: bool
    airview: int


@dataclass
class Services6(AirOSDataClass):
    """Leaf definition."""

    dhcpc: bool
    dhcpd: bool
    pppoe: bool


@dataclass
class Airview6(AirOSDataClass):
    """Leaf definition."""

    enabled: int


@dataclass
class Firewall(AirOSDataClass):
    """Leaf definition."""

    iptables: bool
    ebtables: bool
    ip6tables: bool
    eb6tables: bool


@dataclass
class Throughput(AirOSDataClass):
    """Leaf definition."""

    tx: int
    rx: int


@dataclass
class ServiceTime(AirOSDataClass):
    """Leaf definition."""

    time: int
    link: int


@dataclass
class Polling(AirOSDataClass):
    """Leaf definition."""

    cb_capacity: int
    dl_capacity: int
    ul_capacity: int
    use: int
    tx_use: int
    rx_use: int
    atpc_status: int
    fixed_frame: bool
    gps_sync: bool
    ff_cap_rep: bool
    flex_mode: int | None = None  # Not present in all devices


@dataclass
class Stats(AirOSDataClass):
    """Leaf definition."""

    rx_bytes: int
    rx_packets: int
    rx_pps: int
    tx_bytes: int
    tx_packets: int
    tx_pps: int


@dataclass
class EvmData(AirOSDataClass):
    """Leaf definition."""

    usage: int
    cinr: int
    evm: list[list[int]]


@dataclass
class Airmax(AirOSDataClass):
    """Leaf definition."""

    actual_priority: int
    beam: int
    desired_priority: int
    cb_capacity: int
    dl_capacity: int
    ul_capacity: int
    atpc_status: int
    rx: EvmData
    tx: EvmData


@dataclass
class EthList(AirOSDataClass):
    """Leaf definition."""

    ifname: str
    enabled: bool
    plugged: bool
    duplex: bool
    speed: int
    snr: list[int]
    cable_len: int


@dataclass
class GPSData(AirOSDataClass):
    """Leaf definition."""

    lat: float | int | None = None
    lon: float | int | None = None
    fix: int | None = None
    sats: int | None = None  # LiteAP GPS
    dim: int | None = None  # LiteAP GPS
    dop: float | int | None = None  # LiteAP GPS
    alt: float | int | None = None  # LiteAP GPS
    time_synced: int | None = None  # LiteAP GPS


@dataclass
class UnmsStatus(AirOSDataClass):
    """Leaf definition."""

    status: int
    timestamp: str | None = None


@dataclass
class Remote(AirOSDataClass):
    """Leaf definition."""

    device_id: str
    hostname: str
    platform: str
    version: str
    time: str
    cpuload: float | int | None
    temperature: int
    totalram: int
    freeram: int
    netrole: str
    sys_id: str
    tx_throughput: int
    rx_throughput: int
    uptime: int
    power_time: int
    compat_11n: int
    signal: int
    rssi: int
    noisefloor: int
    tx_power: int
    distance: int  # In meters
    rx_chainmask: int
    chainrssi: list[int]
    tx_ratedata: list[int]
    tx_bytes: int
    rx_bytes: int
    cable_loss: int
    ethlist: list[EthList]
    ipaddr: list[str]
    oob: bool
    unms: UnmsStatus
    airview: int
    service: ServiceTime
    mode: WirelessMode | None = None  # Investigate why remotes can have no mode set
    ip6addr: list[str] | None = None  # For v4 only devices
    height: int | None = None
    age: int | None = None  # At least not present on 8.7.11
    gps: GPSData | None = (
        None  # Reported NanoStation 5AC 8.7.18 without GPS Core 150491
    )
    antenna_gain: int | None = None  # Reported on Prism 6.3.5? and LiteBeam 8.7.8

    @classmethod
    def __pre_deserialize__(cls, d: dict[str, Any]) -> dict[str, Any]:
        """Pre-deserialize hook for Wireless."""
        _check_and_log_unknown_enum_value(d, "mode", WirelessMode, "Remote", "mode")
        return d


@dataclass
class Disconnected(AirOSDataClass):
    """Leaf definition for disconnected devices."""

    mac: str
    lastip: str
    hostname: str
    platform: str
    reason_code: int
    disconnect_duration: int
    airos_connected: bool = False  # Mock add to determine Disconnected vs Station
    signal: int | None = None  # Litebeam 5AC can have no signal


@dataclass
class Station(AirOSDataClass):
    """Leaf definition for connected/active devices."""

    mac: str
    lastip: str
    signal: int
    rssi: int
    noisefloor: int
    chainrssi: list[int]
    tx_idx: int
    rx_idx: int
    tx_nss: int
    rx_nss: int
    tx_latency: int
    distance: int  # In meters
    tx_packets: int
    tx_lretries: int
    tx_sretries: int
    uptime: int
    dl_signal_expect: int
    ul_signal_expect: int
    cb_capacity_expect: int
    dl_capacity_expect: int
    ul_capacity_expect: int
    dl_rate_expect: int
    ul_rate_expect: int
    dl_linkscore: int
    ul_linkscore: int
    dl_avg_linkscore: int
    ul_avg_linkscore: int
    tx_ratedata: list[int]
    stats: Stats
    airmax: Airmax
    last_disc: int
    remote: Remote
    airos_connected: bool = True  # Mock add to determine Disconnected vs Station


@dataclass
class Wireless(AirOSDataClass):
    """Leaf definition."""

    essid: str
    band: int
    compat_11n: int
    hide_essid: int
    apmac: str
    frequency: int
    center1_freq: int
    dfs: int
    distance: int  # In meters
    security: Security
    noisef: int
    txpower: int
    aprepeater: bool
    rstatus: int
    chanbw: int
    rx_chainmask: int
    tx_chainmask: int
    cac_state: int
    cac_timeout: int
    rx_idx: int
    rx_nss: int
    tx_idx: int
    tx_nss: int
    throughput: Throughput
    service: ServiceTime
    polling: Polling
    count: int
    sta: list[Station]
    sta_disconnected: list[Disconnected]
    ieeemode: IeeeMode
    mode: WirelessMode | None = None  # Investigate further (see WirelessMode in Remote)
    nol_state: int | None = None  # Reported on Prism 6.3.5? and LiteBeam 8.7.8
    nol_timeout: int | None = None  # Reported on Prism 6.3.5? and LiteBeam 8.7.8
    antenna_gain: int | None = None  # Reported on Prism 6.3.5? and LiteBeam 8.7.8

    @classmethod
    def __pre_deserialize__(cls, d: dict[str, Any]) -> dict[str, Any]:
        """Pre-deserialize hook for Wireless."""
        _check_and_log_unknown_enum_value(d, "mode", WirelessMode, "Wireless", "mode")
        _check_and_log_unknown_enum_value(
            d, "ieeemode", IeeeMode, "Wireless", "ieeemode"
        )
        _check_and_log_unknown_enum_value(
            d, "security", Security, "Wireless", "security"
        )
        return d


@dataclass
class Wireless6(AirOSDataClass):
    """Leaf definition."""

    essid: str
    hide_essid: int
    apmac: str
    countrycode: int
    channel: int
    frequency: str
    dfs: int
    opmode: str
    antenna: str
    chains: str
    signal: int
    rssi: int
    noisef: int
    txpower: int
    ack: int
    distance: int  # In meters
    ccq: int
    txrate: str
    rxrate: str
    security: Security
    qos: str
    rstatus: int
    cac_nol: int
    nol_chans: int
    wds: int
    aprepeater: int  # Not bool as v8
    chanbw: int
    mode: Wireless6Mode | None = None

    @classmethod
    def __pre_deserialize__(cls, d: dict[str, Any]) -> dict[str, Any]:
        """Pre-deserialize hook for Wireless6."""
        _check_and_log_unknown_enum_value(d, "mode", Wireless6Mode, "Wireless6", "mode")
        _check_and_log_unknown_enum_value(
            d, "security", Security, "Wireless", "security"
        )
        return d


@dataclass
class InterfaceStatus(AirOSDataClass):
    """Leaf definition."""

    plugged: bool
    tx_bytes: int
    rx_bytes: int
    tx_packets: int
    rx_packets: int
    tx_errors: int
    rx_errors: int
    tx_dropped: int
    rx_dropped: int
    ipaddr: str
    speed: int
    duplex: bool
    snr: list[int] | None = None
    cable_len: int | None = None
    ip6addr: list[dict[str, Any]] | None = None


@dataclass
class Interface6Status(AirOSDataClass):
    """Leaf definition."""

    duplex: bool
    plugged: bool
    speed: int
    snr: list[int] | None = None
    cable_len: int | None = None
    ip6addr: list[dict[str, Any]] | None = None


@dataclass
class Interface(AirOSDataClass):
    """Leaf definition."""

    ifname: str
    hwaddr: str
    enabled: bool
    status: InterfaceStatus
    mtu: int


@dataclass
class Interface6(AirOSDataClass):
    """Leaf definition."""

    ifname: str
    hwaddr: str
    enabled: bool
    status: Interface6Status
    mtu: int | None = None  # Reported unpresent on v6


@dataclass
class ProvisioningMode(AirOSDataClass):
    """Leaf definition."""


@dataclass
class NtpClient(AirOSDataClass):
    """Leaf definition."""


@dataclass
class GPSMain(AirOSDataClass):
    """Leaf definition."""

    lat: float | int | None
    lon: float | int | None
    fix: int


@dataclass
class Derived(AirOSDataClass):
    """Contain custom data generated by this module."""

    mac: str  # Base device MAC address (i.e. eth0)
    mac_interface: str  # Interface derived from

    # Split for WirelessMode
    station: bool
    access_point: bool

    # Split for WirelessMode
    ptp: bool
    ptmp: bool

    role: DerivedWirelessRole
    mode: DerivedWirelessMode

    # Lookup of model_id (presumed via SKU)
    sku: str


@dataclass
class AirOS8Data(AirOSDataBaseClass):
    """Dataclass for AirOS v8 devices."""

    chain_names: list[ChainName]
    host: Host
    genuine: str
    services: Services
    firewall: Firewall
    portfw: bool
    wireless: Wireless
    interfaces: list[Interface]
    provmode: Any
    ntpclient: Any
    unms: UnmsStatus
    derived: Derived
    gps: GPSData | None = (
        None  # Reported NanoStation 5AC 8.7.18 without GPS Core 150491
    )


@dataclass
class AirOS6Data(AirOSDataBaseClass):
    """Dataclass for AirOS v6 devices."""

    airview: Airview6
    host: Host6
    genuine: str
    services: Services6
    firewall: Firewall
    wireless: Wireless6
    interfaces: list[Interface6]
    unms: UnmsStatus
    derived: Derived



"""Request tests."""

import logging
from unittest.mock import AsyncMock, MagicMock, patch

import aiohttp
import pytest

from airos.airos8 import AirOS8
from airos.exceptions import (
    AirOSConnectionAuthenticationError,
    AirOSDataMissingError,
    AirOSDeviceConnectionError,
)

# pylint: disable=redefined-outer-name


@pytest.fixture
def mock_session() -> MagicMock:
    """Return a mock aiohttp ClientSession."""
    return MagicMock(spec=aiohttp.ClientSession)


@pytest.fixture
def mock_airos8_device(mock_session: MagicMock) -> AirOS8:
    """Return a mock AirOS instance with string host."""
    return AirOS8(
        host="192.168.1.3",
        username="testuser",
        password="testpassword",
        session=mock_session,
    )


@pytest.mark.asyncio
async def test_request_json_success(
    mock_airos8_device: AirOS8,
    mock_session: MagicMock,
) -> None:
    """Test successful JSON request."""
    expected_response_data = {"key": "value"}
    mock_response = AsyncMock()
    mock_response.status = 200
    mock_response.text = AsyncMock(return_value='{"key": "value"}')
    mock_response.raise_for_status = MagicMock()

    mock_session.request.return_value.__aenter__.return_value = mock_response

    with patch.object(mock_airos8_device, "connected", True):
        response_data = await mock_airos8_device._request_json("GET", "/test/path")  # noqa: SLF001

    assert response_data == expected_response_data
    mock_session.request.assert_called_once()
    mock_session.request.assert_called_once_with(
        "GET",
        "/test/path",
        json=None,
        data=None,
        headers={},
    )


@pytest.mark.asyncio
async def test_request_json_connection_error(
    mock_airos8_device: AirOS8,
    mock_session: MagicMock,
) -> None:
    """Test handling of a connection error."""
    mock_session.request.return_value.__aenter__.side_effect = (
        aiohttp.ClientConnectionError
    )

    with (
        patch.object(mock_airos8_device, "connected", True),
        pytest.raises(AirOSDeviceConnectionError),
    ):
        await mock_airos8_device._request_json("GET", "/test/path")  # noqa: SLF001


@pytest.mark.asyncio
async def test_request_json_http_error(
    mock_airos8_device: AirOS8,
    mock_session: MagicMock,
) -> None:
    """Test handling of a non-200 HTTP status code."""
    mock_response = AsyncMock()
    mock_response.status = 401
    mock_response.raise_for_status = MagicMock(
        side_effect=aiohttp.ClientResponseError(
            request_info=MagicMock(), history=(), status=401, message="Unauthorized"
        )
    )
    mock_response.text = AsyncMock(return_value="{}")

    mock_session.request.return_value.__aenter__.return_value = mock_response

    with (
        patch.object(mock_airos8_device, "connected", True),
        pytest.raises(AirOSConnectionAuthenticationError),
    ):
        await mock_airos8_device._request_json("GET", "/test/path")  # noqa: SLF001

    mock_response.raise_for_status.assert_called_once()


@pytest.mark.asyncio
async def test_request_json_non_json_response(
    mock_airos8_device: AirOS8,
    mock_session: MagicMock,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test handling of a response that is not valid JSON."""
    mock_response = AsyncMock()
    mock_response.status = 200
    mock_response.text = AsyncMock(return_value="NOT-A-JSON-STRING")
    mock_response.raise_for_status = MagicMock()
    mock_session.request.return_value.__aenter__.return_value = mock_response

    with (
        patch.object(mock_airos8_device, "connected", True),
        pytest.raises(AirOSDataMissingError),
        caplog.at_level(logging.DEBUG),
    ):
        await mock_airos8_device._request_json("GET", "/test/path")  # noqa: SLF001

    assert "Failed to decode JSON from /test/path" in caplog.text


@pytest.mark.asyncio
async def test_request_json_with_params_and_data(
    mock_airos8_device: AirOS8,
    mock_session: MagicMock,
) -> None:
    """Test request with parameters and data."""
    mock_response = AsyncMock()
    mock_response.status = 200
    mock_response.text = AsyncMock(return_value="{}")
    mock_response.raise_for_status = MagicMock()

    mock_session.request.return_value.__aenter__.return_value = mock_response

    params = {"param1": "value1"}
    data = {"key": "value"}

    with patch.object(mock_airos8_device, "connected", True):
        await mock_airos8_device._request_json(  # noqa: SLF001
            "POST", "/test/path", json_data=params, form_data=data
        )

    mock_session.request.assert_called_once_with(
        "POST",
        "/test/path",
        json=params,
        data=data,
        headers={},
    )



"""Ubiquiti AirOS base class."""

from __future__ import annotations

from abc import ABC
import asyncio
from collections.abc import Callable
from http.cookies import SimpleCookie
import json
import logging
from typing import Any, Generic, TypeVar
from urllib.parse import urlparse

import aiohttp
from mashumaro.exceptions import InvalidFieldValue, MissingField

from .data import (
    AirOSDataBaseClass,
    DerivedWirelessMode,
    DerivedWirelessRole,
    redact_data_smart,
)
from .exceptions import (
    AirOSConnectionAuthenticationError,
    AirOSConnectionSetupError,
    AirOSDataMissingError,
    AirOSDeviceConnectionError,
    AirOSKeyDataMissingError,
    AirOSMultipleMatchesFoundException,
    AirOSUrlNotFoundError,
)
from .model_map import UispAirOSProductMapper

_LOGGER = logging.getLogger(__name__)

AirOSDataModel = TypeVar("AirOSDataModel", bound=AirOSDataBaseClass)


class AirOS(ABC, Generic[AirOSDataModel]):
    """AirOS connection class."""

    data_model: type[AirOSDataModel]

    def __init__(
        self,
        data_model: type[AirOSDataModel],
        host: str,
        username: str,
        password: str,
        session: aiohttp.ClientSession,
        use_ssl: bool = True,
    ):
        """Initialize AirOS class."""
        self.data_model = data_model
        self.username = username
        self.password = password

        parsed_host = urlparse(host)
        scheme = (
            parsed_host.scheme
            if parsed_host.scheme
            else ("https" if use_ssl else "http")
        )
        hostname = parsed_host.hostname if parsed_host.hostname else host

        self.base_url = f"{scheme}://{hostname}"

        self.session = session

        self._use_json_for_login_post = False
        self._auth_cookie: str | None = None
        self._csrf_id: str | None = None
        self.connected: bool = False
        self.current_csrf_token: str | None = None

        # Mostly 8.x API endpoints, login/status are the same in 6.x
        self._login_urls = {
            "default": f"{self.base_url}/api/auth",
            "v6_alternative": f"{self.base_url}/login.cgi",
        }
        self._status_cgi_url = f"{self.base_url}/status.cgi"
        # Presumed 8.x only endpoints
        self._stakick_cgi_url = f"{self.base_url}/stakick.cgi"
        self._provmode_url = f"{self.base_url}/api/provmode"
        self._warnings_url = f"{self.base_url}/api/warnings"
        self._update_check_url = f"{self.base_url}/api/fw/update-check"
        self._download_url = f"{self.base_url}/api/fw/download"
        self._download_progress_url = f"{self.base_url}/api/fw/download-progress"
        self._install_url = f"{self.base_url}/fwflash.cgi"

    @staticmethod
    def derived_wireless_data(
        derived: dict[str, Any], response: dict[str, Any]
    ) -> dict[str, Any]:
        """Add derived wireless data to the device response."""
        # Access Point / Station vs PTP/PtMP
        wireless_mode = response.get("wireless", {}).get("mode", "")
        match wireless_mode:
            case "ap-ptmp":
                derived["access_point"] = True
                derived["ptmp"] = True
                derived["role"] = DerivedWirelessRole.ACCESS_POINT
                derived["mode"] = DerivedWirelessMode.PTMP
            case "sta-ptmp":
                derived["station"] = True
                derived["ptmp"] = True
                derived["mode"] = DerivedWirelessMode.PTMP
            case "ap-ptp":
                derived["access_point"] = True
                derived["ptp"] = True
                derived["role"] = DerivedWirelessRole.ACCESS_POINT
            case "sta-ptp":
                derived["station"] = True
                derived["ptp"] = True
        return derived

    @staticmethod
    def _derived_data_helper(
        response: dict[str, Any],
        derived_wireless_data_func: Callable[
            [dict[str, Any], dict[str, Any]], dict[str, Any]
        ],
    ) -> dict[str, Any]:
        """Add derived data to the device response."""
        sku: str = "UNKNOWN"

        devmodel = (response.get("host") or {}).get("devmodel", "UNKNOWN")
        try:
            sku = UispAirOSProductMapper().get_sku_by_devmodel(devmodel)
        except KeyError:
            _LOGGER.warning(
                "Unknown SKU/Model ID for %s. Please report at "
                "https://github.com/CoMPaTech/python-airos/issues so we can add support.",
                devmodel,
            )
            sku = "UNKNOWN"
        except AirOSMultipleMatchesFoundException as err:  # pragma: no cover
            _LOGGER.warning(
                "Multiple SKU/Model ID matches found for model '%s': %s. Please report at "
                "https://github.com/CoMPaTech/python-airos/issues so we can add support.",
                devmodel,
                err,
            )
            sku = "AMBIGUOUS"

        derived: dict[str, Any] = {
            "station": False,
            "access_point": False,
            "ptp": False,
            "ptmp": False,
            "role": DerivedWirelessRole.STATION,
            "mode": DerivedWirelessMode.PTP,
            "sku": sku,
        }
        # WIRELESS
        derived = derived_wireless_data_func(derived, response)

        # INTERFACES
        addresses = {}
        interface_order = ["br0", "eth0", "ath0"]

        interfaces = response.get("interfaces", [])

        # No interfaces, no mac, no usability
        if not interfaces:
            _LOGGER.error("Failed to determine interfaces from AirOS data")
            raise AirOSKeyDataMissingError from None

        for interface in interfaces:
            if interface["enabled"]:  # Only consider if enabled
                addresses[interface["ifname"]] = interface["hwaddr"]

        # Fallback take fist alternate interface found
        derived["mac"] = interfaces[0]["hwaddr"]
        derived["mac_interface"] = interfaces[0]["ifname"]

        for interface in interface_order:
            if interface in addresses:
                derived["mac"] = addresses[interface]
                derived["mac_interface"] = interface
                break

        response["derived"] = derived

        return response

    def derived_data(self, response: dict[str, Any]) -> dict[str, Any]:
        """Add derived data to the device response (instance method for polymorphism)."""
        return self._derived_data_helper(response, self.derived_wireless_data)

    def _get_authenticated_headers(
        self,
        ct_json: bool = False,
        ct_form: bool = False,
    ) -> dict[str, str]:
        """Construct headers for an authenticated request."""
        headers = {}
        if ct_json:
            headers["Content-Type"] = "application/json"
        elif ct_form:
            headers["Content-Type"] = "application/x-www-form-urlencoded"

        if self._csrf_id:  # pragma: no cover
            headers["X-CSRF-ID"] = self._csrf_id

        if self._auth_cookie:  # pragma: no cover
            headers["Cookie"] = f"AIROS_{self._auth_cookie}"

        return headers

    def _store_auth_data(self, response: aiohttp.ClientResponse) -> None:
        """Parse the response from a successful login and store auth data."""
        self._csrf_id = response.headers.get("X-CSRF-ID")

        # Parse all Set-Cookie headers to ensure we don't miss AIROS_* cookie
        cookie = SimpleCookie()
        for set_cookie in response.headers.getall("Set-Cookie", []):
            cookie.load(set_cookie)
        for key, morsel in cookie.items():
            if key.startswith("AIROS_"):
                self._auth_cookie = morsel.key[6:] + "=" + morsel.value
                break

    async def _request_json(
        self,
        method: str,
        url: str,
        headers: dict[str, Any] | None = None,
        json_data: dict[str, Any] | None = None,
        form_data: dict[str, Any] | None = None,
        authenticated: bool = False,
        ct_json: bool = False,
        ct_form: bool = False,
    ) -> dict[str, Any] | Any:
        """Make an authenticated API request and return JSON response."""
        # Pass the content type flags to the header builder
        request_headers = (
            self._get_authenticated_headers(ct_json=ct_json, ct_form=ct_form)
            if authenticated
            else {}
        )
        if headers:
            request_headers.update(headers)

        try:
            if url not in self._login_urls.values() and not self.connected:
                _LOGGER.error("Not connected, login first")
                raise AirOSDeviceConnectionError from None

            async with self.session.request(
                method,
                url,
                json=json_data,
                data=form_data,
                headers=request_headers,  # Pass the constructed headers
            ) as response:
                response.raise_for_status()
                response_text = await response.text()
                _LOGGER.debug("Successfully fetched JSON from %s", url)

                # If this is the login request, we need to store the new auth data
                if url in self._login_urls.values():
                    self._store_auth_data(response)
                    self.connected = True

                return json.loads(response_text)
        except aiohttp.ClientResponseError as err:
            _LOGGER.error(
                "Request to %s failed with status %s: %s", url, err.status, err.message
            )
            if err.status in [401, 403]:
                raise AirOSConnectionAuthenticationError from err
            if err.status in [404]:
                raise AirOSUrlNotFoundError from err
            raise AirOSConnectionSetupError from err
        except (TimeoutError, aiohttp.ClientError) as err:
            _LOGGER.exception("Error during API call to %s", url)
            raise AirOSDeviceConnectionError from err
        except json.JSONDecodeError as err:
            _LOGGER.error("Failed to decode JSON from %s", url)
            raise AirOSDataMissingError from err
        except asyncio.CancelledError:
            _LOGGER.warning("Request to %s was cancelled", url)
            raise

    async def login(self) -> None:
        """Login to AirOS device."""
        payload = {"username": self.username, "password": self.password}
        try:
            await self._request_json(
                "POST", self._login_urls["default"], json_data=payload
            )
        except AirOSUrlNotFoundError:
            pass  # Try next URL
        except AirOSConnectionSetupError as err:
            raise AirOSConnectionSetupError("Failed to login to AirOS device") from err
        else:
            return

        try:  # Alternative URL
            await self._request_json(
                "POST",
                self._login_urls["v6_alternative"],
                form_data=payload,
                ct_form=True,
            )
        except AirOSConnectionSetupError as err:
            raise AirOSConnectionSetupError(
                "Failed to login to default and alternate AirOS device urls"
            ) from err

    async def status(self) -> AirOSDataModel:
        """Retrieve status from the device."""
        response = await self._request_json(
            "GET", self._status_cgi_url, authenticated=True
        )

        try:
            adjusted_json = self.derived_data(response)
            return self.data_model.from_dict(adjusted_json)
        except InvalidFieldValue as err:
            # Log with .error() as this is a specific, known type of issue
            redacted_data = redact_data_smart(response)
            _LOGGER.error(
                "Failed to deserialize AirOS data due to an invalid field value: %s",
                redacted_data,
            )
            raise AirOSKeyDataMissingError from err
        except MissingField as err:
            # Log with .exception() for a full stack trace
            redacted_data = redact_data_smart(response)
            _LOGGER.exception(
                "Failed to deserialize AirOS data due to a missing field: %s",
                redacted_data,
            )
            raise AirOSKeyDataMissingError from err

    async def update_check(self, force: bool = False) -> dict[str, Any]:
        """Check for firmware updates."""
        if force:
            return await self._request_json(
                "POST",
                self._update_check_url,
                json_data={"force": True},
                authenticated=True,
                ct_form=True,
            )
        return await self._request_json(
            "POST",
            self._update_check_url,
            json_data={},
            authenticated=True,
            ct_json=True,
        )

    async def stakick(self, mac_address: str | None = None) -> bool:
        """Reconnect client station."""
        if not mac_address:
            _LOGGER.error("Device mac-address missing")
            raise AirOSDataMissingError from None

        payload = {"staif": "ath0", "staid": mac_address.upper()}

        await self._request_json(
            "POST",
            self._stakick_cgi_url,
            form_data=payload,
            ct_form=True,
            authenticated=True,
        )
        return True

    async def provmode(self, active: bool = False) -> bool:
        """Set provisioning mode."""
        action = "stop"
        if active:
            action = "start"

        payload = {"action": action}
        await self._request_json(
            "POST",
            self._provmode_url,
            form_data=payload,
            ct_form=True,
            authenticated=True,
        )
        return True

    async def warnings(self) -> dict[str, Any]:
        """Get warnings."""
        return await self._request_json("GET", self._warnings_url, authenticated=True)

    async def progress(self) -> dict[str, Any]:
        """Get download progress for updates."""
        payload: dict[str, Any] = {}
        return await self._request_json(
            "POST",
            self._download_progress_url,
            json_data=payload,
            ct_json=True,
            authenticated=True,
        )

    async def download(self) -> dict[str, Any]:
        """Download new firmware."""
        payload: dict[str, Any] = {}
        return await self._request_json(
            "POST",
            self._download_url,
            json_data=payload,
            ct_json=True,
            authenticated=True,
        )

    async def install(self) -> dict[str, Any]:
        """Install new firmware."""
        payload: dict[str, Any] = {"do_update": 1}
        return await self._request_json(
            "POST",
            self._install_url,
            json_data=payload,
            ct_json=True,
            authenticated=True,
        )



"""Ubiquiti AirOS firmware helpers."""

from typing import TypedDict

import aiohttp

from .airos6 import AirOS6
from .airos8 import AirOS8
from .exceptions import AirOSKeyDataMissingError


class DetectDeviceData(TypedDict):
    """Container for device data."""

    fw_major: int
    mac: str
    hostname: str


async def async_get_firmware_data(
    host: str,
    username: str,
    password: str,
    session: aiohttp.ClientSession,
    use_ssl: bool = True,
) -> DetectDeviceData:
    """Connect to a device and return the major firmware version."""
    detect: AirOS8 = AirOS8(host, username, password, session, use_ssl)

    await detect.login()
    raw_status = await detect._request_json(  # noqa: SLF001
        "GET",
        detect._status_cgi_url,  # noqa: SLF001
        authenticated=True,
    )

    fw_version = (raw_status.get("host") or {}).get("fwversion")
    if not fw_version:
        raise AirOSKeyDataMissingError("Missing host.fwversion in API response")

    try:
        fw_major = int(fw_version.lstrip("v").split(".", 1)[0])
    except (ValueError, AttributeError) as exc:
        raise AirOSKeyDataMissingError(
            f"Invalid firmware version '{fw_version}'"
        ) from exc

    if fw_major == 6:
        derived_data = AirOS6._derived_data_helper(  # noqa: SLF001
            raw_status, AirOS6.derived_wireless_data
        )
    else:  # Assume AirOS 8 for all other versions
        derived_data = AirOS8._derived_data_helper(  # noqa: SLF001
            raw_status, AirOS8.derived_wireless_data
        )

    # Extract MAC address and hostname from the derived data
    hostname = derived_data.get("host", {}).get("hostname")
    mac = derived_data.get("derived", {}).get("mac")

    if not hostname:  # pragma: no cover
        raise AirOSKeyDataMissingError("Missing hostname")

    if not mac:  # pragma: no cover
        raise AirOSKeyDataMissingError("Missing MAC address")

    return {
        "fw_major": fw_major,
        "mac": mac,
        "hostname": hostname,
    }
